"""Tests for the parsing functionality (PyFilter)."""

import pytest
from cohere_melody import PyFilter, PyFilterOptions


class TestPyFilterWriteDecoded:
    """Tests for PyFilter.write_decoded method."""

    @pytest.fixture
    def cmd3_filter(self):
        """Create a fresh cmd3 filter for each test."""
        return PyFilter.cmd3()

    def test_plain_text(self, cmd3_filter):
        """Test write_decoded with plain text."""
        result = cmd3_filter.write_decoded("Hello world")
        assert result.content == "Hello world"
        assert result.reasoning is None

    def test_thinking_tags(self, cmd3_filter):
        """Test write_decoded with thinking tags."""
        result = cmd3_filter.write_decoded("<|START_THINKING|>This is a thought")
        assert result.reasoning == "This is a thought"
        assert result.content is None

    def test_response_tags(self, cmd3_filter):
        """Test write_decoded with response tags."""
        result = cmd3_filter.write_decoded("<|START_RESPONSE|>Hello")
        assert result.content == "Hello"
        assert result.reasoning is None

    def test_transition_thinking_to_response(self, cmd3_filter):
        """Test transitioning from thinking to response."""
        result = cmd3_filter.write_decoded("<|START_THINKING|>Thinking...")
        assert result.reasoning is not None

        result = cmd3_filter.write_decoded("<|END_THINKING|><|START_RESPONSE|>Response")
        assert result.content is not None
        assert "Response" in result.content


class TestPyFilterFlushPartials:
    """Tests for PyFilter.flush_partials method."""

    def test_flush_with_buffered_content(self):
        """Test flush_partials returns buffered content when chunk_size > 1."""
        f = PyFilter.cmd3(chunk_size=10)
        result = f.write_decoded("<|START_RESPONSE|>Hi")
        assert result.content is None or result.content == ""
        result = f.flush_partials()
        assert result.content is not None
        assert "Hi" in result.content

    def test_flush_empty_filter(self):
        """Test flush_partials on fresh filter returns default result."""
        f = PyFilter.cmd3()
        result = f.flush_partials()
        assert result.content is None
        assert result.reasoning is None

    def test_flush_after_complete_output(self):
        """Test flush_partials after content already emitted."""
        f = PyFilter.cmd3()
        f.write_decoded("<|START_RESPONSE|>Hello")
        result = f.flush_partials()
        assert result is not None


class TestPyFilterStreamingWorkflow:
    """Tests for complete streaming workflows."""

    def test_complete_workflow(self):
        """Test a complete streaming workflow."""
        f = PyFilter.cmd3()
        all_text = []

        tokens = ["<|START_RESPONSE|>", "Hello", " ", "world", "!"]
        for token in tokens:
            result = f.write_decoded(token)
            if result.content is not None:
                all_text.append(result.content)

        result = f.flush_partials()
        if result.content is not None:
            all_text.append(result.content)

        full_text = "".join(all_text)
        assert "Hello world!" in full_text

    def test_thinking_then_response_workflow(self):
        """Test workflow with thinking followed by response."""
        f = PyFilter.cmd3()

        result = f.write_decoded("<|START_THINKING|>This is a")
        assert result.reasoning == "This is a"
        assert result.content is None

        result = f.write_decoded(" plan.<|END_THINKING|>")
        assert result.reasoning == " plan."

        result = f.write_decoded("<|START_RESPONSE|>This is the final response.")
        assert result.content == "This is the final response."
        assert result.reasoning is None


class TestAggregatedResult:
    """Tests for AggregatedResult attributes."""

    @pytest.fixture
    def sample_result(self):
        """Get a sample AggregatedResult from a filter."""
        f = PyFilter.cmd3()
        result = f.write_decoded("<|START_RESPONSE|>Test")
        return result

    def test_has_content_attribute(self, sample_result):
        """Test that AggregatedResult has content attribute."""
        assert hasattr(sample_result, "content")

    def test_has_reasoning_attribute(self, sample_result):
        """Test that AggregatedResult has reasoning attribute."""
        assert hasattr(sample_result, "reasoning")

    def test_thinking_output_has_reasoning(self):
        """Test that thinking output has reasoning set."""
        f = PyFilter.cmd3()
        result = f.write_decoded("<|START_THINKING|>Thought")
        assert result.reasoning is not None
        assert result.content is None

    def test_response_output_has_content(self):
        """Test that response output has content set."""
        f = PyFilter.cmd3()
        result = f.write_decoded("<|START_RESPONSE|>Response")
        assert result.content is not None
        assert result.reasoning is None


class TestPyFilterOptions:
    """Tests for PyFilterOptions builder pattern."""

    def test_cmd3_option(self):
        """Test cmd3() builder method."""
        opts = PyFilterOptions().cmd3()
        f = PyFilter(opts)
        result = f.write_decoded("<|START_RESPONSE|>Hello")
        assert result.content == "Hello"

    def test_cmd4_option(self):
        """Test cmd4() builder method."""
        opts = PyFilterOptions().cmd4()
        f = PyFilter(opts)
        result = f.write_decoded("Hello")
        assert result.content == "Hello"

    def test_rag_option(self):
        """Test rag() builder method."""
        opts = PyFilterOptions().rag()
        f = PyFilter(opts)
        result = f.write_decoded("Hello")
        assert result is not None

    def test_multi_hop_option(self):
        """Test multi_hop() builder method."""
        opts = PyFilterOptions().multi_hop()
        f = PyFilter(opts)
        result = f.write_decoded("Hello")
        assert result is not None

    def test_search_query_option(self):
        """Test search_query() builder method."""
        opts = PyFilterOptions().search_query()
        f = PyFilter(opts)
        result = f.write_decoded("Hello")
        assert result is not None

    def test_with_chunk_size(self):
        """Test with_chunk_size() builder method."""
        opts = PyFilterOptions().cmd3().with_chunk_size(10)
        f = PyFilter(opts)
        result = f.write_decoded("<|START_RESPONSE|>Hi")
        assert result.content is None or result.content == ""
        result = f.flush_partials()
        assert result.content is not None
        assert "Hi" in result.content

    def test_chained_options(self):
        """Test chaining multiple builder methods."""
        opts = (
            PyFilterOptions()
            .cmd3()
            .with_chunk_size(5)
            .with_left_trimmed()
            .with_right_trimmed()
        )
        f = PyFilter(opts)
        assert f is not None

    def test_with_inclusive_stops(self):
        """Test with_inclusive_stops() builder method."""
        opts = PyFilterOptions().cmd3().with_inclusive_stops(["STOP"])
        f = PyFilter(opts)
        assert f is not None

    def test_with_exclusive_stops(self):
        """Test with_exclusive_stops() builder method."""
        opts = PyFilterOptions().cmd3().with_exclusive_stops(["END"])
        f = PyFilter(opts)
        assert f is not None

    def test_stream_tool_actions(self):
        """Test stream_tool_actions() builder method."""
        opts = PyFilterOptions().cmd3().stream_tool_actions()
        f = PyFilter(opts)
        assert f is not None

    def test_stream_processed_params(self):
        """Test stream_processed_params() builder method."""
        opts = PyFilterOptions().cmd3().stream_processed_params()
        f = PyFilter(opts)
        assert f is not None

    def test_stream_non_grounded_answer(self):
        """Test stream_non_grounded_answer() builder method."""
        opts = PyFilterOptions().cmd3().stream_non_grounded_answer()
        f = PyFilter(opts)
        assert f is not None

    def test_remove_token(self):
        """Test remove_token() builder method."""
        opts = PyFilterOptions().cmd3().remove_token("<|START_THINKING|>")
        f = PyFilter(opts)
        assert f is not None

    def test_options_are_immutable(self):
        """Test that builder methods return new instances."""
        opts1 = PyFilterOptions()
        opts2 = opts1.cmd3()
        opts3 = opts2.with_chunk_size(10)
        assert opts1 is not opts2
        assert opts2 is not opts3



import json
import os
from enum import Enum
from os.path import basename, dirname, isdir, join
from typing import Any

import jinja2
import minijinja
import pytest


class Engine(Enum):
    JINJA2 = 1
    MINIJINJA = 2


def render_template_jinja2(template_dir: str, template_name: str, **kwargs: Any) -> str:
    # This jinja environment should match the huggingface settings from here:
    # https://github.com/huggingface/transformers/blob/57278c904c5158999d31a0db8bfcd63360c37b48/src/transformers/utils/chat_template_utils.py#L455-L460
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_dir),
        lstrip_blocks=True,
        trim_blocks=True,
        extensions=["jinja2.ext.loopcontrols"],
    )
    if "cmd3-v1" in template_name:
        kwargs["add_generation_prompt"] = True
    # Overriding tojson with ensure_ascii=False so that tojson doesn't write unicode
    # characters as \uxxxx
    env.filters["tojson"] = lambda *args, **kwargs: json.dumps(
        *args, ensure_ascii=False, sort_keys=False, **kwargs
    )
    env.policies["json.dumps_kwargs"] = {"sort_keys": False}
    template = env.get_template(template_name)

    return template.render(**kwargs)


def render_template_minijinja(
    template_dir: str, template_name: str, **kwargs: Any
) -> str:
    def loader(name: str) -> str:
        content, _, _ = jinja2.FileSystemLoader(template_dir).get_source(
            None,  # type: ignore[arg-type] # Not used by this loader implementation
            name,
        )
        return content

    def tojson(*args: Any, **kwargs: Any) -> str:
        return json.dumps(*args, ensure_ascii=False, sort_keys=False, **kwargs)

    filters = {"tojson": tojson}

    # This jinja environment should match the huggingface settings from here:
    # https://github.com/huggingface/transformers/blob/57278c904c5158999d31a0db8bfcd63360c37b48/src/transformers/utils/chat_template_utils.py#L455-L460
    env = minijinja.Environment(
        loader=loader,
        filters=filters,
        lstrip_blocks=True,
        trim_blocks=True,
    )
    if "cmd3-v1" in template_name:
        kwargs["add_generation_prompt"] = True

    return env.render_template(template_name, **kwargs)


def render_template(
    template_dir: str,
    template_name: str,
    engine: Engine,
    **kwargs: Any,
) -> str:
    if engine is Engine.MINIJINJA:
        return render_template_minijinja(template_dir, template_name, **kwargs)

    return render_template_jinja2(template_dir, template_name, **kwargs)


# read json from file_path
def read_test_data(file_path: str) -> Any:
    with open(file_path) as file:
        return json.load(file)


def get_template_info(template_path: str) -> tuple[str, str]:
    # Get template directory, and the directory's name
    template_dir = dirname(template_path)

    # Get template name
    template_name = basename(template_path)

    return template_dir, template_name


@pytest.mark.parametrize(
    "template_path, test_dir",
    [
        (
            "templates/jinja/cmd3-v1.jinja",
            "jinja_tests/cmd3_v1_hf/chat_merged_template_v1",
        ),
        (
            "templates/jinja/cmd3-v2.jinja",
            "jinja_tests/cmd3_reasoning_hf/chat_merged_template",
        ),
        ("templates/jinja/cmd4-v1.jinja", "jinja_tests/cmd4_v1/chat_template"),
    ],
)
@pytest.mark.parametrize("engine", [Engine.JINJA2, Engine.MINIJINJA])
def test_render_template(template_path: str, test_dir: str, engine: Engine) -> None:
    template_dir, template_name = get_template_info(template_path)

    # get all .json files from the test_dir
    test_files = [
        f
        for f in os.listdir(test_dir)
        # check not isdir instead of isfile so invalid symlinks are caught
        if not isdir(join(test_dir, f)) and f.endswith(".json")
    ]

    print(f"Running {len(test_files)} tests for {template_path}")
    errors = []
    for test_file in test_files:
        test_data = read_test_data(f"{test_dir}/{test_file}")
        rendered = render_template(template_dir, template_name, engine, **test_data)
        with open(f"{test_dir}/{test_file.replace('.json', '.txt')}") as f:
            expected = f.read()
        try:
            assert rendered == expected
        except AssertionError as e:
            print(f"Failed for template: {test_file}")
            e.add_note(f"input: {json.dumps(test_data, indent=2)}")
            e.add_note(f"test_file: {test_file}")
            errors.append(e)
    if errors:
        raise ExceptionGroup("Tests failed", errors)



from __future__ import annotations

class PyFilterOptions:
    def __init__(self) -> None: ...
    def cmd3(self) -> PyFilterOptions: ...
    def cmd4(self) -> PyFilterOptions: ...
    def rag(self) -> PyFilterOptions: ...
    def multi_hop(self) -> PyFilterOptions: ...
    def search_query(self) -> PyFilterOptions: ...
    def with_chunk_size(self, size: int) -> PyFilterOptions: ...
    def with_left_trimmed(self) -> PyFilterOptions: ...
    def with_right_trimmed(self) -> PyFilterOptions: ...
    def with_inclusive_stops(self, stops: list[str]) -> PyFilterOptions: ...
    def with_exclusive_stops(self, stops: list[str]) -> PyFilterOptions: ...
    def stream_tool_actions(self) -> PyFilterOptions: ...
    def stream_processed_params(self) -> PyFilterOptions: ...
    def stream_non_grounded_answer(self) -> PyFilterOptions: ...
    def no_tools(self) -> PyFilterOptions: ...
    def remove_token(self, token: str) -> PyFilterOptions: ...

class Source:
    tool_call_index: int
    tool_result_indices: list[int]

class FilterCitation:
    start_index: int
    end_index: int
    text: str
    sources: list[Source]
    is_thinking: bool

class SearchQueryDelta:
    index: int
    text: str

class AccumulatedToolCall:
    index: int
    id: str
    name: str
    arguments: str

class FilterAggregatedResult:
    content: str | None
    reasoning: str | None
    tool_calls: list[AccumulatedToolCall]
    citations: list[FilterCitation]
    search_queries: list[SearchQueryDelta]

class PyFilter:
    def __init__(self, options: PyFilterOptions) -> None: ...
    @staticmethod
    def cmd3(chunk_size: int | None = None) -> PyFilter: ...
    @staticmethod
    def cmd4(chunk_size: int | None = None) -> PyFilter: ...
    def write_decoded(self, decoded_token: str) -> FilterAggregatedResult: ...
    def flush_partials(self) -> FilterAggregatedResult: ...
    def process_full(self, token_strings: list[str]) -> FilterAggregatedResult: ...
    def process_full_text(self, text: str) -> FilterAggregatedResult: ...

def render_cmd3(config: dict) -> str: ...
def render_cmd4(config: dict) -> str: ...



"""vLLM integration for *melody*.

Wraps the melody functionality into vLLM parsers for reasoning and tool calls.
"""

from typing import Optional, Sequence, Union, TYPE_CHECKING
from vllm.entrypoints.openai.protocol import (
    ChatCompletionRequest,
    ResponsesRequest,
    DeltaMessage,
    DeltaToolCall,
    DeltaFunctionCall,
    ExtractedToolCallInformation,
    FunctionCall,
    ToolCall,
)
from vllm.reasoning import ReasoningParser, ReasoningParserManager
from vllm.tool_parsers import ToolParser, ToolParserManager
from vllm.transformers_utils.tokenizer import AnyTokenizer

try:
    from cohere_melody import PyFilter, PyFilterOptions  # type: ignore

except ModuleNotFoundError:
    raise RuntimeError("The compiled melody bindings are not available.")

REPLACEMENT_CHAR = "\ufffd"


class CohereCommand2ReasoningParser(ReasoningParser):

    def __init__(self, tokenizer: AnyTokenizer, *args, **kwargs):
        super().__init__(tokenizer, *args, **kwargs)
        self.melody = PyFilter(PyFilterOptions().cmd3())

    def extract_reasoning_streaming(
        self,
        previous_text: str,
        current_text: str,
        delta_text: str,
        previous_token_ids: Sequence[int],
        current_token_ids: Sequence[int],
        delta_token_ids: Sequence[int],
    ) -> Union[DeltaMessage, None]:

        result = self.melody.write_decoded(delta_text)

        content = result.content
        reasoning_content = result.reasoning
        delta_tool_calls = [
            DeltaToolCall(
                id=tc.id,
                index=tc.index,
                type="function",
                function=DeltaFunctionCall(
                    name=tc.name,
                    arguments=tc.arguments,
                ),
            )
            for tc in result.tool_calls
        ]

        if content is None and reasoning_content is None and len(delta_tool_calls) == 0:
            return None

        msg = DeltaMessage()
        if content is not None:
            msg.content = content
        if reasoning_content is not None:
            msg.reasoning_content = reasoning_content
        if len(delta_tool_calls) > 0:
            msg.tool_calls = delta_tool_calls

        return msg

    def extract_reasoning(
        self, model_output: str, request: ChatCompletionRequest | ResponsesRequest
    ) -> tuple[Optional[str], Optional[str]]:
        melody = PyFilter(
            PyFilterOptions()
            .cmd3()
            .remove_token("<|START_ACTION|>")
            .remove_token("<|END_ACTION|>")
        )
        result = melody.process_full_text(model_output)
        return result.reasoning, result.content

    def extract_content_ids(self, input_ids: list[int]) -> list[int]:
        melody = PyFilter(
            PyFilterOptions()
            .cmd3()
            .remove_token("<|START_ACTION|>")
            .remove_token("<|END_ACTION|>")
        )
        token_buf = []
        content_ids = []
        for t in input_ids:
            token_buf.append(t)
            token_str = self.model_tokenizer.decode(
                token_buf, skip_special_tokens=False
            )
            # buffer tokens that generate incomplete strings
            if token_str.endswith(REPLACEMENT_CHAR):
                continue

            result = melody.write_decoded(token_str)
            if result.content is not None:
                content_ids.extend(token_buf)

            token_buf = []
        return content_ids

    def is_reasoning_end(self, input_ids: list[int]) -> bool:
        end_token_id = self.model_tokenizer.convert_tokens_to_ids("<|END_THINKING|>")
        return any(input_id == end_token_id for input_id in reversed(input_ids))


if not TYPE_CHECKING:
    ReasoningParserManager.register_module(["cohere2"])(CohereCommand2ReasoningParser)


class CohereCommand2ToolParser(ToolParser):

    def __init__(self, tokenizer: AnyTokenizer):
        super().__init__(tokenizer)
        self.melody = PyFilter(PyFilterOptions().cmd3())

    def adjust_request(self, request: ChatCompletionRequest) -> ChatCompletionRequest:
        request = super().adjust_request(request)
        request.skip_special_tokens = False
        return request

    def extract_tool_calls_streaming(
        self,
        previous_text: str,
        current_text: str,
        delta_text: str,
        previous_token_ids: Sequence[int],
        current_token_ids: Sequence[int],
        delta_token_ids: Sequence[int],
        request: ChatCompletionRequest,
    ) -> Union[DeltaMessage, None]:

        result = self.melody.write_decoded(delta_text)

        if not result.tool_calls:
            return None

        delta_tool_calls = [
            DeltaToolCall(
                id=tc.id,
                index=tc.index,
                type="function",
                function=DeltaFunctionCall(
                    name=tc.name,
                    arguments=tc.arguments,
                ),
            )
            for tc in result.tool_calls
        ]

        return DeltaMessage(tool_calls=delta_tool_calls)

    def extract_tool_calls(
        self,
        model_output: str,
        request: ChatCompletionRequest,
    ) -> ExtractedToolCallInformation:
        result = self.melody.process_full_text(model_output)

        tool_calls = [
            ToolCall(
                id=tc.id,
                type="function",
                function=FunctionCall(name=tc.name, arguments=tc.arguments),
            )
            for tc in result.tool_calls
        ]

        return ExtractedToolCallInformation(
            tools_called=len(tool_calls) > 0,
            tool_calls=tool_calls,
            content=result.content,
        )


if not TYPE_CHECKING:
    ToolParserManager.register_module(["cohere2"])(CohereCommand2ToolParser)



"""Tests for the templating functionality (render_cmd3, render_cmd4)."""

import pytest
from cohere_melody import render_cmd3, render_cmd4


class TestRenderCmd3:
    """Tests for the render_cmd3 function."""

    def test_empty_messages(self):
        """Test rendering with empty messages."""
        result = render_cmd3({"messages": []})
        assert isinstance(result, str)

    def test_single_user_message(self):
        """Test rendering with a single user message."""
        result = render_cmd3(
            {
                "messages": [
                    {"role": "user", "content": [{"type": "text", "text": "Hello!"}]}
                ]
            }
        )
        assert "USER" in result
        assert "Hello!" in result

    def test_user_and_assistant_messages(self):
        """Test rendering with user and assistant messages."""
        result = render_cmd3(
            {
                "messages": [
                    {"role": "user", "content": [{"type": "text", "text": "Hi"}]},
                    {
                        "role": "chatbot",
                        "content": [{"type": "text", "text": "Hello!"}],
                    },
                ]
            }
        )
        assert "USER" in result
        assert "CHATBOT" in result

    def test_with_dev_instruction(self):
        """Test rendering with developer instruction."""
        result = render_cmd3(
            {
                "messages": [],
                "dev_instruction": "You are a helpful assistant.",
            }
        )
        assert "helpful assistant" in result

    def test_with_tools(self):
        """Test rendering with available tools."""
        result = render_cmd3(
            {
                "messages": [],
                "available_tools": [
                    {
                        "name": "search",
                        "description": "Search the web",
                        "parameters": {"type": "object", "properties": {}},
                    }
                ],
            }
        )
        assert "search" in result

    def test_with_documents(self):
        """Test rendering with documents."""
        result = render_cmd3(
            {
                "messages": [],
                "documents": [{"title": "Doc 1", "content": "Some content"}],
            }
        )
        assert "Doc 1" in result or "content" in result

    def test_safety_mode_strict(self):
        """Test that strict safety mode includes strict safety preamble."""
        result = render_cmd3(
            {
                "messages": [],
                "safety_mode": "strict",
            }
        )
        assert "You are in strict safety mode" in result
        assert "You are in contextual safety mode" not in result

    def test_safety_mode_contextual(self):
        """Test that contextual safety mode includes contextual safety preamble."""
        result = render_cmd3(
            {
                "messages": [],
                "safety_mode": "contextual",
            }
        )
        assert "You are in contextual safety mode" in result
        assert "You are in strict safety mode" not in result

    def test_safety_mode_none(self):
        """Test that none safety mode omits all safety preambles."""
        result = render_cmd3(
            {
                "messages": [],
                "safety_mode": "none",
            }
        )
        assert "You are in strict safety mode" not in result
        assert "You are in contextual safety mode" not in result

    def test_safety_modes_produce_different_outputs(self):
        """Test that different safety modes produce different outputs."""
        results = {}
        for mode in ["none", "strict", "contextual"]:
            results[mode] = render_cmd3(
                {
                    "messages": [],
                    "safety_mode": mode,
                }
            )
        assert results["none"] != results["strict"]
        assert results["none"] != results["contextual"]
        assert results["strict"] != results["contextual"]

    def test_citation_quality_on_includes_grounding(self):
        """Test that citation_quality 'on' includes grounding instructions."""
        result = render_cmd3(
            {
                "messages": [],
                "citation_quality": "on",
                "available_tools": [
                    {
                        "name": "search",
                        "description": "Search the web",
                        "parameters": {"type": "object", "properties": {}},
                    }
                ],
            }
        )
        assert "Grounding" in result

    def test_citation_quality_off_excludes_grounding(self):
        """Test that citation_quality 'off' excludes grounding instructions."""
        result = render_cmd3(
            {
                "messages": [],
                "citation_quality": "off",
                "available_tools": [
                    {
                        "name": "search",
                        "description": "Search the web",
                        "parameters": {"type": "object", "properties": {}},
                    }
                ],
            }
        )
        assert "Grounding" not in result

    def test_citation_quality_produces_different_outputs(self):
        """Test that different citation quality settings produce different outputs."""
        config_base = {
            "messages": [],
            "available_tools": [
                {
                    "name": "search",
                    "description": "Search the web",
                    "parameters": {"type": "object", "properties": {}},
                }
            ],
        }
        result_on = render_cmd3({**config_base, "citation_quality": "on"})
        result_off = render_cmd3({**config_base, "citation_quality": "off"})
        assert result_on != result_off

    def test_with_json_mode(self):
        """Test rendering with JSON mode enabled."""
        result = render_cmd3(
            {
                "messages": [],
                "json_mode": True,
            }
        )
        assert isinstance(result, str)

    def test_with_json_schema(self):
        """Test rendering with JSON schema."""
        result = render_cmd3(
            {
                "messages": [],
                "json_schema": '{"type": "object"}',
            }
        )
        assert isinstance(result, str)

    def test_with_response_prefix(self):
        """Test rendering with response prefix."""
        result = render_cmd3(
            {
                "messages": [],
                "response_prefix": "Sure, here is",
            }
        )
        assert isinstance(result, str)

    def test_invalid_config_type_raises_type_error(self):
        """Test that passing non-dict raises TypeError."""
        with pytest.raises(TypeError):
            render_cmd3("not a dict")

    def test_unknown_field_raises_value_error(self):
        """Test that unknown field raises ValueError."""
        with pytest.raises(ValueError):
            render_cmd3({"messages": [], "unknown_field": "value"})

    def test_invalid_role_raises_value_error(self):
        """Test that invalid role raises ValueError."""
        with pytest.raises(ValueError):
            render_cmd3(
                {
                    "messages": [
                        {
                            "role": "invalid_role",
                            "content": [{"type": "text", "text": "Hi"}],
                        }
                    ]
                }
            )

    def test_invalid_content_type_raises_value_error(self):
        """Test that invalid content type raises ValueError."""
        with pytest.raises(ValueError):
            render_cmd3(
                {
                    "messages": [
                        {
                            "role": "user",
                            "content": [{"type": "invalid_type", "text": "Hi"}],
                        }
                    ]
                }
            )


class TestRenderCmd4:
    """Tests for the render_cmd4 function."""

    def test_empty_messages(self):
        """Test rendering with empty messages."""
        result = render_cmd4({"messages": []})
        assert isinstance(result, str)

    def test_single_user_message(self):
        """Test rendering with a single user message."""
        result = render_cmd4(
            {
                "messages": [
                    {"role": "user", "content": [{"type": "text", "text": "Hello!"}]}
                ]
            }
        )
        assert "USER" in result
        assert "Hello!" in result

    def test_user_and_assistant_messages(self):
        """Test rendering with user and assistant messages."""
        result = render_cmd4(
            {
                "messages": [
                    {"role": "user", "content": [{"type": "text", "text": "Hi"}]},
                    {
                        "role": "chatbot",
                        "content": [{"type": "text", "text": "Hello!"}],
                    },
                ]
            }
        )
        assert "USER" in result
        assert "CHATBOT" in result

    def test_with_dev_instruction(self):
        """Test rendering with developer instruction."""
        result = render_cmd4(
            {
                "messages": [],
                "dev_instruction": "You are a helpful assistant.",
            }
        )
        assert isinstance(result, str)

    def test_with_platform_instruction(self):
        """Test rendering with platform instruction."""
        result = render_cmd4(
            {
                "messages": [],
                "platform_instruction": "Custom platform instruction",
            }
        )
        assert isinstance(result, str)

    def test_with_tools(self):
        """Test rendering with available tools."""
        result = render_cmd4(
            {
                "messages": [],
                "available_tools": [
                    {
                        "name": "calculator",
                        "description": "Perform calculations",
                        "parameters": {"type": "object", "properties": {}},
                    }
                ],
            }
        )
        assert "calculator" in result

    def test_with_documents(self):
        """Test rendering with documents."""
        result = render_cmd4(
            {
                "messages": [],
                "documents": [{"title": "Document", "text": "Content here"}],
            }
        )
        assert isinstance(result, str)

    def test_grounding_enabled_includes_grounding_preamble(self):
        """Test that grounding 'enabled' includes grounding instructions."""
        result = render_cmd4(
            {
                "messages": [],
                "grounding": "enabled",
                "available_tools": [
                    {
                        "name": "search",
                        "description": "Search the web",
                        "parameters": {"type": "object", "properties": {}},
                    }
                ],
            }
        )
        assert "responses and reflections can be grounded" in result

    def test_grounding_disabled_excludes_grounding_preamble(self):
        """Test that grounding 'disabled' excludes grounding instructions."""
        result = render_cmd4(
            {
                "messages": [],
                "grounding": "disabled",
                "available_tools": [
                    {
                        "name": "search",
                        "description": "Search the web",
                        "parameters": {"type": "object", "properties": {}},
                    }
                ],
            }
        )
        assert "responses and reflections can be grounded" not in result

    def test_grounding_produces_different_outputs(self):
        """Test that different grounding settings produce different outputs."""
        config_base = {
            "messages": [],
            "available_tools": [
                {
                    "name": "search",
                    "description": "Search the web",
                    "parameters": {"type": "object", "properties": {}},
                }
            ],
        }
        result_enabled = render_cmd4({**config_base, "grounding": "enabled"})
        result_disabled = render_cmd4({**config_base, "grounding": "disabled"})
        assert result_enabled != result_disabled

    def test_with_json_mode(self):
        """Test rendering with JSON mode enabled."""
        result = render_cmd4(
            {
                "messages": [],
                "json_mode": True,
            }
        )
        assert isinstance(result, str)

    def test_with_json_schema(self):
        """Test rendering with JSON schema."""
        result = render_cmd4(
            {
                "messages": [],
                "json_schema": '{"type": "object"}',
            }
        )
        assert isinstance(result, str)

    def test_with_response_prefix(self):
        """Test rendering with response prefix."""
        result = render_cmd4(
            {
                "messages": [],
                "response_prefix": "Here is my response:",
            }
        )
        assert isinstance(result, str)

    def test_invalid_config_type_raises_type_error(self):
        """Test that passing non-dict raises TypeError."""
        with pytest.raises(TypeError):
            render_cmd4("not a dict")

    def test_unknown_field_raises_value_error(self):
        """Test that unknown field raises ValueError."""
        with pytest.raises(ValueError):
            render_cmd4({"messages": [], "unknown_field": "value"})

    def test_invalid_role_raises_value_error(self):
        """Test that invalid role raises ValueError."""
        with pytest.raises(ValueError):
            render_cmd4(
                {
                    "messages": [
                        {
                            "role": "invalid_role",
                            "content": [{"type": "text", "text": "Hi"}],
                        }
                    ]
                }
            )

    def test_invalid_grounding_raises_value_error(self):
        """Test that invalid grounding value raises ValueError."""
        with pytest.raises(ValueError):
            render_cmd4({"messages": [], "grounding": "invalid"})



import json
import os
from os.path import isfile, join
from typing import Any

import pytest

from test_jinja import Engine, get_template_info, read_test_data, render_template


def get_all_files(test_dir: str) -> list[str]:
    return [
        f
        for f in os.listdir(test_dir)
        if isfile(join(test_dir, f)) and f.endswith(".json")
    ]


def get_test_tuples(
    template_to_test_files: dict[str, list[str]], test_dir: str
) -> list[tuple[str, str, str]]:
    template_test_files: list[tuple[str, str, str]] = []
    for template, test_files in template_to_test_files.items():
        nt = len(test_files)
        template_test_files.extend(zip([template] * nt, [test_dir] * nt, test_files))
    return template_test_files


def get_tests(template: str, test_dir: str) -> list[tuple[str, str, str]]:
    test_files = get_all_files(test_dir)
    template_to_test_files = {template: test_files}
    return get_test_tuples(template_to_test_files, test_dir)


def get_cmd3_v2_tests() -> list[tuple[str, str, str]]:
    # get all .json files from the test_dir
    template_to_test_files = {}
    test_dir = "liquid_tests/rag/cmd3-v2"
    all_test_files = get_all_files(test_dir)

    merged_template = "templates/jinja/cmd3-v2.jinja"
    template_to_test_files[merged_template] = all_test_files

    return get_test_tuples(template_to_test_files, test_dir)


def get_cmd3_v1_tests() -> list[tuple[str, str, str]]:
    # get all .json files from the test_dir
    merged_template = "templates/jinja/cmd3-v1.jinja"

    chat_test_dir = "liquid_tests/chat/cmd3-v1"
    rag_test_dir = "liquid_tests/rag/cmd3-v1"

    return get_tests(merged_template, chat_test_dir) + get_tests(
        merged_template, rag_test_dir
    )


def get_cmd3_v3_tests() -> list[tuple[str, str, str]]:
    # get all .json files from the test_dir
    merged_template = "templates/jinja/cmd3-v3.jinja"

    test_dir = "liquid_tests/rag/cmd3-v3"

    return get_tests(merged_template, test_dir)


def get_cmd4_v1_tests() -> list[tuple[str, str, str]]:
    # get all .json files from the test_dir
    chat_template = "templates/jinja/cmd4-v1.jinja"
    rag_test_dir = "liquid_tests/rag/cmd4-v1"
    return get_tests(chat_template, rag_test_dir)


def get_template_test_files() -> list[tuple[str, str, str]]:
    cmd3_v2_tests = get_cmd3_v2_tests()
    cmd3_v1_tests = get_cmd3_v1_tests()
    cmd3_v3_tests = get_cmd3_v3_tests()
    cmd4_v1_tests = get_cmd4_v1_tests()
    return [*cmd3_v2_tests, *cmd3_v1_tests, *cmd3_v3_tests, *cmd4_v1_tests]


template_test_files = get_template_test_files()


# This test is to run liquid template tests against the jinja template
@pytest.mark.parametrize("template_path, test_dir, test_files", template_test_files)
@pytest.mark.parametrize("engine", [Engine.JINJA2, Engine.MINIJINJA])
def test_render_template(  # noqa: C901
    template_path: str,
    test_dir: str,
    test_files: str | list[str],
    engine: Engine,
) -> None:
    template_dir, template_name = get_template_info(template_path)

    if isinstance(test_files, str):
        test_files = [test_files]
    errors = []
    for test_file in test_files:
        print("Running test: ", test_file)
        test_data = read_test_data(f"{test_dir}/{test_file}")
        test_data = test_data["variables"]

        test_data["bos_token"] = "<BOS_TOKEN>"
        test_data["add_generation_prompt"] = True

        test_data["enable_citations"] = True
        test_data["regen_tool_call_ids"] = False
        if test_data.get("citation_mode") == "OFF":
            test_data["enable_citations"] = False
        if test_data.get("skip_thinking") is None:
            test_data["skip_thinking"] = False
        test_data["reasoning"] = test_data.get("reasoning_options", {}).get(
            "enabled", False
        )
        if test_data.get("json_mode"):
            test_data["response_format"] = {"type": "json_object"}
            json_schema = test_data.get("json_schema")
            if json_schema:
                test_data["response_format"]["schema"] = json_schema
        avail_tools = test_data.get("available_tools")
        if avail_tools:
            tools = []
            for avail_tool in avail_tools:
                defn = avail_tool.get(
                    "definition", {"description": "", "json_schema": "{}"}
                )
                tool: dict[str, Any] = {"type": "function"}
                tool["function"] = {
                    "name": avail_tool.get("name", ""),
                    "description": defn["description"],
                    "parameters": json.loads(defn["json_schema"]),
                }
                tools.append(tool)
            test_data["tools"] = tools
        role_map = {
            "User": "user",
            "Chatbot": "assistant",
            "Tool": "tool",
        }
        docs = test_data.get("documents", [])
        for idx in range(len(docs)):
            docs[idx] = json.loads(docs[idx])

        messages = test_data["messages"]
        preamble = test_data.get("preamble")
        if preamble:
            messages.insert(0, {"role": "system", "content": preamble})
        new_msgs = []
        for message in messages:
            message["role"] = role_map.get(message["role"], message["role"])
            if "content" in message:
                if isinstance(message["content"], str):
                    continue
                for idx, content in enumerate(message["content"]):
                    if (
                        content["type"] == "thinking"
                        and content.get("thinking") is None
                    ):
                        content["thinking"] = content.get("data", "")
                    if content["type"] == "text" and content.get("text") is None:
                        if idx == 0 and "tool_calls" in message:
                            content["type"] = "thinking"
                            content["thinking"] = content.get("data", "")
                        else:
                            content["text"] = content.get("data", "")
            try:
                if "tool_calls" in message:
                    msg_tool_calls = message["tool_calls"]
                    for idx in range(len(msg_tool_calls)):
                        otcall = json.loads(msg_tool_calls[idx])
                        tcall = {
                            "id": otcall["tool_call_id"],
                            "type": "function",
                            "function": {
                                "name": otcall["tool_name"],
                                "arguments": otcall["parameters"],
                            },
                        }
                        msg_tool_calls[idx] = tcall
                if "tool_results" in message:
                    if "content" not in message:
                        message["content"] = []
                    tool_call_id_to_msg = {}
                    if message["tool_results"]:
                        res_call_id = str(message["tool_results"][0]["tool_call_id"])
                        tool_call_id_to_msg[res_call_id] = message
                        message["tool_call_id"] = res_call_id
                    for res in message["tool_results"]:
                        res_call_id = str(res["tool_call_id"])
                        if res_call_id not in tool_call_id_to_msg:
                            # We need to add a new message for this tool result
                            new_msg = {
                                "role": "tool",
                                "tool_call_id": res_call_id,
                                "content": [],
                            }
                            new_msgs.append((new_msg, message))
                            tool_call_id_to_msg[res_call_id] = new_msg
                        cur_msg = tool_call_id_to_msg[res_call_id]
                        for doc in res["documents"]:
                            cur_msg["content"].append(json.loads(doc))

            except TypeError:
                print(f"Failed for {test_file}")
                raise
        for msg, orig_msg in new_msgs[::-1]:
            orig_msg_idx = messages.index(orig_msg)
            messages.insert(orig_msg_idx + 1, msg)

        # print(json.dumps(test_data, indent=2))
        rendered = render_template(template_dir, template_name, engine, **test_data)
        with open(f"{test_dir}/{test_file.replace('.json', '.txt')}") as f:
            expected = f.read()
        # work around for strange difference where liquid logic is inserting space
        # around brackets
        # Ticket to fix in linear: https://linear.app/cohereai/issue/PTS-8620/fix-liquid-template-spacing
        expected = expected.replace('{ "', '{"').replace('" }', '"}')
        try:
            assert rendered == expected
        except AssertionError as e:
            print(f"Failed for template: {test_file}")
            e.add_note(f"input: {json.dumps(test_data, indent=2)}")
            e.add_note(f"test_file: {test_file}")
            errors.append(e)
    if errors:
        raise ExceptionGroup("Tests failed", errors)
